use argon2::{
    password_hash::{rand_core::OsRng, SaltString},
    Argon2, PasswordHash, PasswordHasher, PasswordVerifier,
};
use rusqlite::Row;
use serde::{Deserialize, Serialize};

use super::{DbError, DbPool};

#[derive(Serialize, Deserialize, Clone)]
pub struct User {
    pub id: i64,
    pub username: String,
    pub password_hash: String,
    pub admin: bool,
}

#[derive(Serialize, Deserialize)]
pub struct NewUser {
    pub username: String,
    pub password: String,
    pub admin: bool,
}

impl User {
    pub fn from_row(row: &Row<'_>) -> Result<Self, rusqlite::Error> {
        Ok(Self {
            id: row.get("id")?,
            username: row.get("username")?,
            password_hash: row.get("password_hash")?,
            admin: row.get("admin")?,
        })
    }

    pub fn by_username(pool: &DbPool, username: impl AsRef<str>) -> Result<Option<Self>, DbError> {
        let conn = pool.get()?;

        let mut stmt = conn.prepare("select * from users where username = $1")?;

        match stmt.query_row((username.as_ref(),), User::from_row) {
            Ok(user) => Ok(Some(user)),
            Err(rusqlite::Error::QueryReturnedNoRows) => Ok(None),
            Err(err) => Err(DbError::Db(err)),
        }
    }

    pub fn verify_password(&self, password: impl AsRef<str>) -> bool {
        let password_hash = PasswordHash::new(&self.password_hash).unwrap();

        Argon2::default()
            .verify_password(password.as_ref().as_bytes(), &password_hash)
            .is_ok()
    }

    pub fn all(pool: &DbPool) -> Result<Vec<Self>, DbError> {
        let conn = pool.get()?;

        let mut stmt = conn.prepare("select * from users")?;
        let users: Result<Vec<_>, _> = stmt.query_map((), Self::from_row)?.collect();

        Ok(users?)
    }

    pub fn remove_by_username(pool: &DbPool, username: impl AsRef<str>) -> Result<bool, DbError> {
        let conn = pool.get()?;
        let mut stmt = conn.prepare("delete from users where username = $1")?;

        Ok(stmt.execute((username.as_ref(),))? > 0)
    }
}

fn hash_password(password: impl AsRef<str>) -> String {
    let salt = SaltString::generate(&mut OsRng);
    let argon2 = Argon2::default();

    argon2
        .hash_password(password.as_ref().as_bytes(), &salt)
        .unwrap()
        .to_string()
}

impl NewUser {
    pub fn insert(self, pool: &DbPool) -> Result<User, DbError> {
        let conn = pool.get()?;

        let password_hash = hash_password(&self.password);

        let mut stmt = conn.prepare(
            "insert into users (username, password_hash, admin) values ($1, $2, $3) returning *",
        )?;

        Ok(stmt.query_row((&self.username, password_hash, &self.admin), User::from_row)?)
    }
}
