mod cli;
mod db;
mod server;

use std::{fs, path::Path, sync::Arc};

use clap::Parser;
use cli::UserCmd;
use db::DbError;
use r2d2_sqlite::SqliteConnectionManager;
use tera::Tera;
use tower_http::compression::CompressionLayer;

const MIGRATIONS: [&str; 5] = [
    include_str!("migrations/000_initial.sql"),
    include_str!("migrations/001_plants.sql"),
    include_str!("migrations/002_comments.sql"),
    include_str!("migrations/003_events.sql"),
    include_str!("migrations/004_auth.sql"),
];
const DB_FILENAME: &str = "db.sqlite3";

#[derive(Clone)]
pub struct Context {
    pool: db::DbPool,
    tera: Arc<Tera>,
}

fn run_user_cli(data_dir: impl AsRef<Path>, cmd: UserCmd) -> Result<(), DbError> {
    let manager = SqliteConnectionManager::file(data_dir.as_ref().join(DB_FILENAME));
    let pool = r2d2::Pool::new(manager)?;

    match cmd.cmd {
        cli::UserSubCmd::List => {
            let users = db::User::all(&pool)?;

            println!("id\tusername\tis admin");

            for user in users {
                println!("{}\t{}\t{}", user.id, user.username, user.admin);
            }
        }
        cli::UserSubCmd::Add {
            username,
            password,
            admin,
        } => {
            db::NewUser {
                username,
                password,
                admin,
            }
            .insert(&pool)?;
        }
        cli::UserSubCmd::Remove { username } => {
            db::User::remove_by_username(&pool, &username)?;
        }
    }

    Ok(())
}

#[tokio::main]
async fn main() {
    let args = cli::Cli::parse();

    match args.cmd {
        cli::Subcommands::Serve { domain, port } => {
            tracing_subscriber::fmt::init();

            if !fs::exists(&args.data_dir).unwrap() {
                fs::create_dir_all(&args.data_dir).unwrap();
            }

            let manager = SqliteConnectionManager::file(args.data_dir.join(DB_FILENAME));
            let pool = r2d2::Pool::new(manager).unwrap();
            db::run_migrations(&pool, &MIGRATIONS).unwrap();

            let tera =
                Tera::new(&format!("{}/**/*", args.templates_dir.to_string_lossy())).unwrap();

            let ctx = Context {
                pool,
                tera: Arc::new(tera),
            };
            let app = server::app(ctx, &args.static_dir)
                .layer(CompressionLayer::new().br(true).gzip(true));

            let address = format!("{}:{}", domain, port);

            tracing::info!("Starting server on {address}");

            let listener = tokio::net::TcpListener::bind(address).await.unwrap();
            axum::serve(listener, app.into_make_service())
                .await
                .unwrap();
        }
        cli::Subcommands::User(cmd) => {
            run_user_cli(&args.data_dir, cmd).unwrap();
        }
    }
}
